/*
        accessor.h -- access dynamic data members
        (C) Copyright 1995  John Webster Small
        All rights reserved

        TYPE must be a selectible class having default
        and copy initializer constructors, overloaded
        assignment operator, and overloaded stream
        insertion and extraction operators.
*/

#ifndef accessor_h
#define accessor_h

/* LINTLIBRARY */

#ifndef assert_h
  #include <assert.h>
  #ifndef assert_h
    #define assert_h
  #endif
#endif

#ifndef iostream_h
  #include <iostream.h>
  #ifndef iostream_h
    #define iostream_h
  #endif
#endif

#ifndef iomanip_h
  #include <iomanip.h>
  #ifndef iomanip_h
    #define iomanip_h
  #endif
#endif

template <class TYPE>
class Accessor
{
    size_t _arrayCells;
    TYPE * _instance;
    void deepCopy(const Accessor<TYPE >& a);
    static TYPE * _newDefault(size_t& arrayCells)
         { arrayCells = 0; return new TYPE; }
  public:
    Accessor() : _instance(0), _arrayCells(0) {}
    Accessor(TYPE * instance, size_t arrayCells = 0)
        : _instance(instance), _arrayCells(arrayCells) {}
    Accessor(const Accessor<TYPE >& a)
        : _instance(0), _arrayCells(0)
        { deepCopy(a); }
    ~Accessor() { release(); }
    Accessor<TYPE >& operator=(const Accessor<TYPE >& a)
        { deepCopy(a); return *this; }
    TYPE * replace
        (TYPE * newInstance, size_t arrayCells_ = 0);
    TYPE * operator()(TYPE * newInstance)
        { return replace(newInstance,0); }
    TYPE * operator()
        (TYPE * newInstance, size_t arrayCells_)
        { return replace(newInstance,arrayCells_); }
    TYPE * ready (TYPE * (*newDefault)(size_t& arrayCells)
        = _newDefault);
    int ok() { return _instance != 0; }
    void release();
    operator TYPE * () { return ready(); }
    size_t arrayCells() { return _arrayCells; }
    TYPE * operator->()
        { assert(_instance); return _instance; }
    TYPE& operator*()
        { assert(_instance); return *_instance; }
    TYPE& operator[](size_t n)
        { assert(_instance && (!n || n < _arrayCells));
          return _instance[n]; }
    ostream& storeOn(ostream& os);
    istream& readFrom(istream& is);
    friend ostream& operator<<
        (ostream& os, Accessor<TYPE >& a)
        { return a.storeOn(os); }
    friend istream& operator>>
        (istream& is, Accessor<TYPE >& a)
        { return a.readFrom(is); }
};

template <class TYPE>
void Accessor<TYPE >:: deepCopy
    (const Accessor<TYPE >& a)
{
    release();
    if (a._arrayCells)  {
        _instance = new TYPE[a._arrayCells];
        if (_instance)
            for (size_t i = 0; i < a._arrayCells; i++)
                 _instance[i] = a._instance[i];
    }
    else
        _instance = new TYPE(*a._instance);
}

template <class TYPE>
TYPE *
Accessor<TYPE >:: replace
    (TYPE * newInstance, size_t arrayCells_)
{
    release();
    _instance = newInstance;
    _arrayCells = arrayCells_;
    return _instance;
}

template <class TYPE>
TYPE *
Accessor<TYPE >:: ready
    (TYPE * (*newDefault)(size_t& arrayCells))
{
    if (!_instance && newDefault)
        _instance = (*newDefault)(_arrayCells);
    return _instance;
}

template <class TYPE>
void
Accessor<TYPE >:: release()
{
    if (_instance) if (_arrayCells)
        delete[] _instance;
    else
        delete _instance;
    _instance = 0;
    _arrayCells = 0;
}

template <class TYPE>
ostream& Accessor<TYPE >:: storeOn(ostream& os)
{
    int i = (_instance? 1 : 0);
    if ((os << i << endl) && i)
      if (os << _arrayCells << '\n')
        if (_arrayCells)
            for (size_t j = 0; j < _arrayCells; j++)
                os << _instance[j] << endl;
        else
            os << *_instance << endl;
    return os;
}

template <class TYPE>
istream& Accessor<TYPE >:: readFrom(istream& is)
{
    release();
    int i;

    is >> i; is.get();
    if (is && i) if (is >> _arrayCells)
        if (_arrayCells)  {
            _instance = new TYPE[_arrayCells];
            if (_instance)
                for (size_t j = 0; j < _arrayCells; j++)
                    (is >> _instance[j]).get();
        }
        else  {
            _instance = new TYPE;
            if (_instance)
               (is >> *_instance).get();
        }
    return is;
}

#endif
